
//	dm.c

#include "dm.h"
int	vb= 0;			// verbose
int	ms;			// ms-stereo
int	gmin[CH];		// average amp for qp
double	bitrate= BR;		// bitrate [kbps]
double	b2= .1/BR/BR;		// .1/(bitrate*bitrate)
double	bitsample= (BR/44.1/CH);// bitrate per sample
int	ibf[CH][L];		// input ring buffer
int	obf[CH][L];		// output ring buffer
int	rip;			// readin pointer to ibf
int	prp;			// process pointer to ibf
int	diff[CH][BL];		// predicted difference
int	coef[CH][BL];		// DCT coefficients
int	qcoef[CH][BL];		// quantized coefficient
int	rcoef[CH][BL];		// reconstructed coeff
int	qm[BL];			// quantize matrix
int	qp[CH];			// quantize parameter
int	buff=BR*1000;		// time constant 1sec
FILE	*in, *bs, *outfp;	// file pointers
int	pw[C];			// power
int	mt[C];			// m table
int	mtp;			// m table pointer
#ifdef HT
#define N HTL
int	ht= 0;
int	hist[256][256];		// histgram[lev][run+1]
int	total;			// total of histgram
int	r[N], l[N], h[N];	// run level hist
int	co[N], le[N];		// cod, len
int	more[N],less[N];
FILE	*fp;

/*
**	pf() and huffman()
*/
void
pf(char *c, int *le)
{
	int i;
	fprintf(fp,"int\n%s[%d]={",c,N);
	for(i=0;i<N;i++){ if(i%8==0)fprintf(fp,"\n"); fprintf(fp,"%6d,", le[i]); }
	fprintf(fp,"\n};\n");
}

void
huffman()
{
	int Max= 0x7fffffff,k,m,i,j,min,i0,i1,tmp,c;
	int max=0,mi=0,ml=0;
	if((fp=fopen("hc","wb"))==NULL){fprintf(stderr,"can't open hc\n");exit(1);}

	tmp= 0;
	for(j=0;j<N-1;j++){
		for(max=k=0;k<256;k++) for(i=0;i<256;i++)
			if(hist[k][i]>max){ max= hist[k][i]; mi= i; ml= k; }
		r[j]= mi;
		l[j]= ml;
		tmp+= h[j]= hist[ml][mi];
		hist[ml][mi]= 0;
	}
	tmp= total-tmp;
	for(j=N-2;j>=0;j--) // position escape code
		if(h[j]<tmp){ h[j+1]= h[j]; r[j+1]= r[j]; l[j+1]= l[j]; }//down them
		else { r[j+1]= 1; l[j+1]= -1; h[j+1]= tmp; break; }//insert escape code
	fprintf(fp,"int esc= %d;\n",j+1);
	for(i=0;i<N;i++)if(l[i]==0)break;
	fprintf(fp,"int eob= %d;\n",i);
	pf("run",r); pf("level",l);

	for(j=0;j<N;j++) more[j]= less[j]= co[j]= le[j]= 0;

	for(j=0;j<N-1;j++){
		i0= 0; min= Max; // find minimum
		for(i=0;i<N;i++) if(min>=h[i]){ min=h[i]; i0= i; }
		tmp= min; h[i0]= Max;
		i1= 0; min= Max; // find next minimum
		for(i=0;i<N;i++) if(min>=h[i]){ min=h[i]; i1= i; }
		h[i1]+= tmp;
		more[j]= i1;
		less[j]= i0;
	}
	for(j=N-2;j>=0;j--){
		m= more[j];
		k= less[j];
		le[k]= le[m]= le[m]+1;
		c= co[m];
		if(m<k) { co[k]= c*2; co[m]= c*2+1; }
		 else   { co[m]= c*2; co[k]= c*2+1; }
	}
	pf("cod",co); pf("len",le);
}
#endif

/*
**	output codebit with specified value and length to "bs" file pointer
**	c: code itself, b0: code length (0<=b0<=32)
**	int	ecnt: number of residue bits as LSBs in ebf
**	Ui	 ebf: 32 bit encoder buffer
**

|<-------  32 ----------->|
	|xxxxxxxxxxxxxxxxx|
<-- i ->|<-----ecnt ----->|	code
			  |yyyyyyyyyyyyyyyyyyyy|
			  |<-------- bit ----->|
	|<-------------32-----------><-- b-i ->|
				     yyyyyyyyyy|
				     <--ecnt-->

**
*/
int	ecnt= 0;
Ui	ebf=  0;

int
code_out()
{
	if(ecnt){
		ebf <<= 32-ecnt;
		putc(ebf>>24,bs);
		if(ecnt> 8)putc(ebf>>16,bs);
		if(ecnt>16)putc(ebf>>8, bs);
		if(ecnt>24)putc(ebf,    bs);
		}
	return (ebf= ecnt= 0);
}

int
code(Ui c,int b0)
{
	int	b;
	Ui	e=ebf;
	b= b0;
	if(b+ecnt>=32){
		int i= 32-ecnt;
		b-= i; ecnt= b;
		if(i==32)e= c;
		  else   e= e<<i | c>>b;
		ebf= c;
		putc(e>>24,bs); putc(e>>16,bs); putc(e>>8,bs); putc(e,bs);
		return b0;
		}
	ecnt+= b;
	ebf= e << b | c;
	return b0;
}

/*
**	Read into ring buffer ibf[CH][L].
**	(1) 2 bytes are read from file pointer in, 
**	(2) make it an int value and get leaky differetial.
**	(3) save it into a ring buffer ibf[CH][index &(L-1)].
*/
int
readin()
{
	static int x,y;
	int c,i,t,d,e;
	int sum= 0;

	for(i=0;i< ((prp-rip-BL/2)&L1) ;i++){
		int k= (rip+i)&L1;
		if((c= getc(in))==EOF) break;
		t= (short)(getc(in)<<8|c);
		d= ibf[0][k]= t - x; x= t - t/LEAK;

		c= getc(in);
		t= (short)(getc(in)<<8|c);
		e= ibf[1][k]= t - y; y= t - t/LEAK;

		sum+= (d*d+e*e)/AL;
		if((k&(AL-1))==AL-1) { pw[k/AL]= sum; sum= 0; }
		}
	rip= (rip+i)&L1;
	return i;
}


//	quantize
void
quant(ch,n)
{
	int i;

	qp[ch]= buff*gmin[ch]*CH*b2;
	if(qp[ch]>MAXQ) qp[ch]= MAXQ;
	if(qp[ch]<MINQ) qp[ch]= MINQ;
	qp[ch]= qp[ch]/MINQ*MINQ;

	if(vb){
		if(ch==0) fprintf(stderr,"ms=%d ", ms);
		fprintf(stderr,"buff=%d ", buff);
		fprintf(stderr,"q=%d ",qp[ch]/MINQ);
	}

	for(i=0;i<n;i++){
		int c= coef[ch][i];
		int q= qp[ch]*qm[i];

		if(c<0) q= (c*QV-q/2)/q;
		else	q= (c*QV+q/2)/q;

		qcoef[ch][i]= q;
		rcoef[ch][i]= q*qp[ch]*qm[i]/QV;
	}
}
/*
**		n
**	1xx	1,2,3,4
**	01xxx	5,..,12,
**	001xxxx 13,.,28
*/
int
lvlc(n)
{
	int k, bits= 0;
	if(n<5) bits+= code(3+n, 3);
	else {
		n+=3; // 8..
		for(k=4; n>>k; k++); // k= 4..
		bits+= code(0, k-3);
		bits+= code(n, k);
		}
	return bits;
}

//	runlev()
int
runlev(ru,q)
{
	int t,a,bits= 0;
	if(q<0) a= -q;
	  else  a= q;

	for(t=0;t<HTL;t++) if(ru==run[t] && a==level[t]) break;
	if(t<HTL) bits+= code(cod[t],len[t]);	// 2D VLC
	 else {
		t= esc;
		bits+= code(cod[t],len[t]);	// escape
		bits+= code(ru,BLL);		// + run
		bits+= lvlc(a);			// + level
		}
	bits+= code(q<0,1);	// sign bit
#ifdef HT
	if(ru<256&&a<256) hist[a][ru]++; total++;
#endif
	return bits;
}


//	codes()
void
codes(ch, n)
{
	int i,ru= 0,q,bits= 0,k;

	if(ch==0) bits+= code(ms,1);	// ms bit
	bits+= code(qp[ch]/MINQ,12);	// quant scale

	for(i=0;i<n;i++){
		k= i;
#ifdef HT
		hist[0][1]++; total++;
#endif
		ru++;
		if((q= qcoef[ch][k])!=0){ bits+= runlev(ru,q); ru= 0; }
		}
	bits+= code(cod[eob],len[eob]);
#ifdef HT
	hist[0][1]++; total++;
#endif
	if(vb){
		if(ch)fprintf(stderr,"bt=%d\n",bits);
		 else fprintf(stderr,"bt=%d, ",bits);
		}
	buff+= bits - n*bitsample; // coded_bits - blocksize * bit/sample
}

void
datain(l)
{
	int i;
	for(i=0;i<l;i++){
		int k= (prp+i)&L1;
		diff[0][i]= ibf[0][k];
		diff[1][i]= ibf[1][k];
	}
	prp= (prp+l)&L1;
}

/*
**	proc(): process a frame (by m-value and next m-value)
**	m (m-value): 0=full frame(BL), 1=half frame, 2=quater frame.
**	nm: next m-value (m of the next frame).
**	n: frame length, BL: full frame size
*/
void
proc(m,nm)
{
	static int lm;
	int n= BL>>m;

	buff+= code(nm,3);
	datain(n);
	if(vb) fprintf(stderr,"m=%d: ",m);
	fmdct(ibf,coef,lm,m,nm);
	quant(0,n); codes(0,n);// L
	quant(1,n); codes(1,n);// R
	imdct(rcoef,obf,lm,m,nm);
	lm= m;
}

/*
**	mkmt(): make m-table from frame length
**	length is divided to 2's powers
**
**	example: 7--> 1, 2, 4
**	M=1 1-> m=0
**	M=2 1-> m=1, 2->m=0
**	M=3 1-> m=2, 2->m=1, 4->m=0
*/
void
mkmt(len)
{
	int i;
	for(i=0;i<M;i++)
		if(len&(1<<i))
			mt[(mtp++)&C1]= (M-1)-i;
}
/*
**	analize(): analize framing and process frames
**	if length before k is BL/AL or the block's power is larger than
**	the avearged power*AT, then there should be a frame whose length is k-m.
**	avarage power is calculated by a= (a + pw[k&C1]) / 2;
**
**	mt: ring buffer which holds m-values.
**	lmp: last mt ring pointer (saving last mtp-1 )
**	mtp: mt ring pointer (incremented by mkmt())
**	last processed frame number is mtp-2.
**	The frames from lmp= mtp-1 to new mtp-2 be processed.
*/
void
analize()
{
	static unsigned int m,k,lmp,sf,a;
	int i,l,ke= rip/AL;
	do {
		int p= pw[k&C1];
		if(k-m == BL/AL || (k&& p>a*AT) || (sf&& p<a)) {
			mkmt(k-m); m= k; sf= p>a*AT;
		}
		a= (a+p)/2;
	} while ((++k&C1)!= ke);

	l= (mtp-1-lmp)&C1;
	for(i=lmp;i<lmp+l;i++) proc(mt[i&C1], mt[(i+1)&C1]);
	lmp= mtp-1;
}

/*
**	usage()
*/
void
usage(char *c)
{
	fprintf(stderr,
"Usage: %s [-r N] [file.wav [bitstream]]\n -r N: bitrate (def. %d [kbps])\n -v V: volume(0-100(def. 80))\n - : reconst wav file to stdout(def. audio)\n",c,BR);
	exit(1);
}

/*
**	main()
*/
int
main(int argc,char ** argv)
{
	int i,l,c,vol= 80;
	char fn[256]= ".dms",*cp= fn, *cmd= *argv;
	in=  stdin; outfp= stdout;
	init_ct();
	argc--; argv++;
	while(argc){
		if(strcmp(*argv,"-r")==0) {
			argv++; argc--;
			bitrate= atof(*argv);
			bitsample= bitrate/44.1/CH;
			buff= bitrate*1000;
			b2= .1/(bitrate*bitrate);
		}
		else if(strcmp(*argv,"-v")==0) {
			argv++; argc--;
			vol= atoi(*argv);
		}
		else if(strcmp(*argv,"-")==0) audio= 0;
		else if(**argv=='-') usage(cmd);
		else break;
		argc--; argv++;
	}
	if(argc >2) usage(cmd);
	if(argc) {
		strcpy(fn,*argv++);
		if(!(in= fopen(fn,"rb")))
			{ fprintf(stderr,"can't open %s\n",fn); exit(1); }
		if(argc==2) strcpy(fn,*argv);
		else {
			for(i=strlen(fn)-1;i>=0;i--)
				if(fn[i]=='/'){ cp= fn+i+1; break; }
			if(cp[strlen(cp)-4]=='.')
				cp[strlen(cp)-4]= 0;
			strcat(cp,".dms");
		}
	}
	if(!(bs= fopen(cp,"wb")))
		{ fprintf(stderr,"can't create %s\n",cp); exit(1); }

	if(audio) aopen(2,44100,vol);
	for(i=0;i<44;i++)
		{ c= getc(in); if(audio==0)putc(c,outfp); buff+= code(c,8); }

	while((l= readin())!=0) analize();
	code_out();
	if(audio)aclose();
#ifdef HT
	if(ht) huffman();
#endif
	return 0;
}
